"use strict";
// Copyright (c) Jupyter Development Team.
// Distributed under the terms of the Modified BSD License.
Object.defineProperty(exports, "__esModule", { value: true });
const apputils_1 = require("@jupyterlab/apputils");
const coreutils_1 = require("@jupyterlab/coreutils");
const ui_components_1 = require("@jupyterlab/ui-components");
const algorithm_1 = require("@lumino/algorithm");
const signaling_1 = require("@lumino/signaling");
const widgets_1 = require("@lumino/widgets");
const contents_1 = require("./contents");
/**
 * The base url for a mybinder deployment.
 */
const MY_BINDER_BASE_URL = 'https://mybinder.org/v2/gh';
/**
 * The className for disabling the mybinder button.
 */
const MY_BINDER_DISABLED = 'jp-MyBinderButton-disabled';
/**
 * Widget for hosting the GitHub filebrowser.
 */
class GitHubFileBrowser extends widgets_1.Widget {
    constructor(browser, drive) {
        super();
        this._binderActive = false;
        this._changeGuard = false;
        this.addClass('jp-GitHubBrowser');
        this.layout = new widgets_1.PanelLayout();
        this.layout.addWidget(browser);
        this._browser = browser;
        this._drive = drive;
        // Create an editable name for the user/org name.
        this.userName = new GitHubUserInput();
        this.userName.node.title = 'Click to edit user/organization';
        this._browser.toolbar.addItem('user', this.userName);
        this.userName.nameChanged.connect(this._onUserChanged, this);
        // Create a button that opens GitHub at the appropriate
        // repo+directory.
        this._openGitHubButton = new apputils_1.ToolbarButton({
            onClick: () => {
                let url = this._drive.baseUrl;
                // If there is no valid user, open the GitHub homepage.
                if (!this._drive.validUser) {
                    window.open(url);
                    return;
                }
                const localPath = this._browser.model.manager.services.contents.localPath(this._browser.model.path);
                const resource = contents_1.parsePath(localPath);
                url = coreutils_1.URLExt.join(url, resource.user);
                if (resource.repository) {
                    url = coreutils_1.URLExt.join(url, resource.repository, 'tree', 'master', resource.path);
                }
                window.open(url);
            },
            iconClass: 'jp-GitHub-icon jp-Icon jp-Icon-16',
            tooltip: 'Open this repository on GitHub'
        });
        this._openGitHubButton.addClass('jp-GitHub-toolbar-item');
        this._browser.toolbar.addItem('GitHub', this._openGitHubButton);
        // Create a button the opens MyBinder to the appropriate repo.
        this._launchBinderButton = new apputils_1.ToolbarButton({
            onClick: () => {
                // If binder is not active for this directory, do nothing.
                if (!this._binderActive) {
                    return;
                }
                const localPath = this._browser.model.manager.services.contents.localPath(this._browser.model.path);
                const resource = contents_1.parsePath(localPath);
                const url = coreutils_1.URLExt.join(MY_BINDER_BASE_URL, resource.user, resource.repository, 'master');
                // Attempt to open using the JupyterLab tree handler
                const tree = coreutils_1.URLExt.join('lab', 'tree', resource.path);
                window.open(url + `?urlpath=${tree}`);
            },
            tooltip: 'Launch this repository on mybinder.org',
            iconClass: 'jp-MyBinderButton jp-Icon jp-Icon-16'
        });
        this._launchBinderButton.addClass('jp-GitHub-toolbar-item');
        this._browser.toolbar.addItem('binder', this._launchBinderButton);
        // Add our own refresh button, since the other one is hidden
        // via CSS.
        let refresher = new apputils_1.ToolbarButton({
            icon: ui_components_1.refreshIcon,
            onClick: () => {
                this._browser.model.refresh();
            },
            tooltip: 'Refresh File List'
        });
        refresher.addClass('jp-GitHub-toolbar-item');
        this._browser.toolbar.addItem('gh-refresher', refresher);
        // Set up a listener to check if we can launch mybinder.
        this._browser.model.pathChanged.connect(this._onPathChanged, this);
        // Trigger an initial pathChanged to check for binder state.
        this._onPathChanged();
        this._drive.rateLimitedState.changed.connect(this._updateErrorPanel, this);
    }
    /**
     * React to a change in user.
     */
    _onUserChanged() {
        if (this._changeGuard) {
            return;
        }
        this._changeGuard = true;
        this._browser.model.cd(`/${this.userName.name}`).then(() => {
            this._changeGuard = false;
            this._updateErrorPanel();
            // Once we have the new listing, maybe give the file listing
            // focus. Once the input element is removed, the active element
            // appears to revert to document.body. If the user has subsequently
            // focused another element, don't focus the browser listing.
            if (document.activeElement === document.body) {
                const listing = this._browser.layout.widgets[2];
                listing.node.focus();
            }
        });
    }
    /**
     * React to the path changing for the browser.
     */
    _onPathChanged() {
        const localPath = this._browser.model.manager.services.contents.localPath(this._browser.model.path);
        const resource = contents_1.parsePath(localPath);
        // If we are not already changing the user name, set it.
        if (!this._changeGuard) {
            this._changeGuard = true;
            this.userName.name = resource.user;
            this._changeGuard = false;
            this._updateErrorPanel();
        }
        // Check for a valid user.
        if (!this._drive.validUser) {
            this._launchBinderButton.addClass(MY_BINDER_DISABLED);
            this._binderActive = false;
            return;
        }
        // Check for a valid repo.
        if (!resource.repository) {
            this._launchBinderButton.addClass(MY_BINDER_DISABLED);
            this._binderActive = false;
            return;
        }
        // If we are in the root of the repository, check for one of
        // the special files indicating we can launch the repository on mybinder.
        // TODO: If the user navigates to a subdirectory without hitting the root
        // of the repository, we will not check for whether the repo is binder-able.
        // Figure out some way around this.
        if (resource.path === '') {
            const item = algorithm_1.find(this._browser.model.items(), i => {
                return (i.name === 'requirements.txt' ||
                    i.name === 'environment.yml' ||
                    i.name === 'apt.txt' ||
                    i.name === 'REQUIRE' ||
                    i.name === 'Dockerfile' ||
                    (i.name === 'binder' && i.type === 'directory'));
            });
            if (item) {
                this._launchBinderButton.removeClass(MY_BINDER_DISABLED);
                this._binderActive = true;
                return;
            }
            else {
                this._launchBinderButton.addClass(MY_BINDER_DISABLED);
                this._binderActive = false;
                return;
            }
        }
        // If we got this far, we are in a subdirectory of a valid
        // repository, and should not change the binderActive status.
        return;
    }
    /**
     * React to a change in the validity of the drive.
     */
    _updateErrorPanel() {
        const localPath = this._browser.model.manager.services.contents.localPath(this._browser.model.path);
        const resource = contents_1.parsePath(localPath);
        const rateLimited = this._drive.rateLimitedState.get();
        const validUser = this._drive.validUser;
        // If we currently have an error panel, remove it.
        if (this._errorPanel) {
            const listing = this._browser.layout.widgets[2];
            listing.node.removeChild(this._errorPanel.node);
            this._errorPanel.dispose();
            this._errorPanel = null;
        }
        // If we are being rate limited, make an error panel.
        if (rateLimited) {
            this._errorPanel = new GitHubErrorPanel('You have been rate limited by GitHub! ' +
                'You will need to wait about an hour before ' +
                'continuing');
            const listing = this._browser.layout.widgets[2];
            listing.node.appendChild(this._errorPanel.node);
            return;
        }
        // If we have an invalid user, make an error panel.
        if (!validUser) {
            const message = resource.user
                ? `"${resource.user}" appears to be an invalid user name!`
                : 'Please enter a GitHub user name';
            this._errorPanel = new GitHubErrorPanel(message);
            const listing = this._browser.layout.widgets[2];
            listing.node.appendChild(this._errorPanel.node);
            return;
        }
    }
}
exports.GitHubFileBrowser = GitHubFileBrowser;
/**
 * A widget that hosts an editable field,
 * used to host the currently active GitHub
 * user name.
 */
class GitHubUserInput extends widgets_1.Widget {
    constructor() {
        super();
        this._name = '';
        this._nameChanged = new signaling_1.Signal(this);
        this.addClass('jp-GitHubUserInput');
        const layout = (this.layout = new widgets_1.PanelLayout());
        const wrapper = new widgets_1.Widget();
        wrapper.addClass('jp-GitHubUserInput-wrapper');
        this._input = document.createElement('input');
        this._input.placeholder = 'GitHub User';
        this._input.className = 'jp-GitHubUserInput-input';
        wrapper.node.appendChild(this._input);
        layout.addWidget(wrapper);
    }
    /**
     * The current name of the field.
     */
    get name() {
        return this._name;
    }
    set name(value) {
        if (value === this._name) {
            return;
        }
        const old = this._name;
        this._name = value;
        this._input.value = value;
        this._nameChanged.emit({
            oldValue: old,
            newValue: value
        });
    }
    /**
     * A signal for when the name changes.
     */
    get nameChanged() {
        return this._nameChanged;
    }
    /**
     * Handle the DOM events for the widget.
     *
     * @param event - The DOM event sent to the widget.
     *
     * #### Notes
     * This method implements the DOM `EventListener` interface and is
     * called in response to events on the main area widget's node. It should
     * not be called directly by user code.
     */
    handleEvent(event) {
        switch (event.type) {
            case 'keydown':
                switch (event.keyCode) {
                    case 13: // Enter
                        event.stopPropagation();
                        event.preventDefault();
                        this.name = this._input.value;
                        this._input.blur();
                        break;
                    default:
                        break;
                }
                break;
            case 'blur':
                event.stopPropagation();
                event.preventDefault();
                this.name = this._input.value;
                break;
            case 'focus':
                event.stopPropagation();
                event.preventDefault();
                this._input.select();
                break;
            default:
                break;
        }
    }
    /**
     * Handle `after-attach` messages for the widget.
     */
    onAfterAttach(msg) {
        this._input.addEventListener('keydown', this);
        this._input.addEventListener('blur', this);
        this._input.addEventListener('focus', this);
    }
    /**
     * Handle `before-detach` messages for the widget.
     */
    onBeforeDetach(msg) {
        this._input.removeEventListener('keydown', this);
        this._input.removeEventListener('blur', this);
        this._input.removeEventListener('focus', this);
    }
}
exports.GitHubUserInput = GitHubUserInput;
/**
 * A widget hosting an error panel for the browser,
 * used if there is an invalid user name or if we
 * are being rate-limited.
 */
class GitHubErrorPanel extends widgets_1.Widget {
    constructor(message) {
        super();
        this.addClass('jp-GitHubErrorPanel');
        const image = document.createElement('div');
        const text = document.createElement('div');
        image.className = 'jp-GitHubErrorImage';
        text.className = 'jp-GitHubErrorText';
        text.textContent = message;
        this.node.appendChild(image);
        this.node.appendChild(text);
    }
}
exports.GitHubErrorPanel = GitHubErrorPanel;
